import numpy as np
import soundfile as sf
from scipy.linalg import toeplitz
import matplotlib.pyplot as plt
from matplotlib.ticker import MaxNLocator

# Read in audiofiles
wn_noise, samplerate = sf.read('lp_noise.wav')
colored, samplerate = sf.read('filtered.wav')
l = len(wn_noise)

# Set length of prediction filter
N = 5

# Abortion criteria
epsilon = 0.00001
E_thr = 0 # 0.001, 0.01

# Auto-correlation estimation
corr_long = np.convolve(colored,np.flip(colored))
corr = corr_long[l-1:]/l

#############################
# Levinson-Durbin Recursion #
#############################

# Prepare some variables
# Initilization
h_opt = np.zeros((1,))
E = np.zeros((N,))
E_init = corr[0]

# Iterate over filter length
for i in range(N):

    # First coefficient
    if i == 0:
        h_opt[i] = corr[1]/corr[0]
        E[i] = E_init*(1-h_opt**2)

    # Recursion
    else:
        # Save results of last iteration
        h_opt_old = h_opt
        # Prepare for new iteration
        h_opt = np.zeros((i+1,))

        # Reversed correlation vector
        r_tilde = np.flip(corr[1:i+1])
        # New coefficient
        h_opt[i] = (corr[i+1] - np.inner(r_tilde,h_opt_old))/(corr[0]-np.inner(r_tilde,h_opt_old[::-1]))

        # Update old coefficients
        h_opt[0:i] = h_opt_old - np.dot(h_opt[i],h_opt_old[::-1])

        # Error power
        E[i] = E[i-1]*(1-h_opt[i]**2)

        # Numerical problems
        if h_opt[i]**2 > 1 + epsilon:
            h_opt = np.zeros((N,))
            h_opt[0:i] = h_opt_old
            break

        # Convergence
        if 10*np.log10(E[i-1]/E[i]) < E_thr:
            h_opt = np.append(h_opt, np.zeros(N-i-1,))
            break

# Plot normalized error power (gain) over iteration
ax = plt.figure().gca()
ax.plot(range(0,N+1),10*np.log10(np.append(E_init,E)/E_init))
ax.xaxis.set_major_locator(MaxNLocator(integer=True))
plt.xlabel('Iteration / Prediction order')
plt.ylabel('Error power reduction in dB')
plt.grid()
plt.show(block='True')

# Prediction-error filter
h_e = np.append(1, -h_opt)

# Error signal
error_long = np.convolve(colored,h_e)
error = error_long[0:l]

# Compare error powers
print('Estimated error power: ',E[i])
print('Actual error power: ', np.inner(error,error)/l)

############################
# Basic Inversion Solution #
############################

# Vector 
r = corr[1:N+1]

# Matrix
R = toeplitz(corr[0:N])

# Optimal filter coefficients (prediction)
hopt = np.matmul(np.linalg.inv(R),r)

# Compare filter coefficients 
ax = plt.figure().gca()
ax.stem(hopt, label='Inversion', use_line_collection=True)
ax.plot(h_opt, label='Recursion', c='green')
ax.xaxis.set_major_locator(MaxNLocator(integer=True))
plt.xlabel('Coefficient index')
plt.ylabel('Value')
plt.legend()
plt.grid()
plt.show(block='True')